/* Copyright 1997 Jim McBeath
 *
 * Jim McBeath, January 12, 1997
 */

package jimmc.roots;

import java.io.*;
import java.util.Dictionary;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

/** CommandInterpreter is a simple command interpreter for a RootsBase.
 */
public class CommandInterpreter implements Runnable {
	RootsBase base;
	LineNumberReader inputStream;
	public boolean runMe;
	String prompt = "> ";
	public Hashtable commandDictionary;
	public Hashtable printFormatDictionary;		/* print formats */
	public Hashtable recordFormatDictionary;	/* record formats */

	/** Create a CommandInterpreter with no attached RootsBase */
	public CommandInterpreter() {
		/* Set up our table of commands */
		commandDictionary = new Hashtable();
		Command.AddDefaultCommands(this);

		/* Set up our table of print formats */
		printFormatDictionary = new Hashtable();
		PrintFormat.AddDefaultFormats(this);

		/* Set up our table of record formats */
		recordFormatDictionary = new Hashtable();
		RecordFormat.AddDefaultFormats(this);
	}
	/** Create a CommandInterpreter with an attached RootsBase */
	public CommandInterpreter(RootsBase b) {
		this();
		setRootsBase(b);
	}

	/** Add one command to our dictionary */
	public void addCommand(Command c) {
		commandDictionary.put(c.getName(),c);
	}
	/** Get a command by name. */
	public Command findCommand(String name) {
		Command c = (Command)(commandDictionary.get(name));
		return c;
	}

	/** Add one print format to our dictionary */
	public void addPrintFormat(PrintFormat f) {
		printFormatDictionary.put(f.getName(),f);
	}
	/** Get a print format by name. */
	public PrintFormat findPrintFormat(String name) {
		PrintFormat f = (PrintFormat)(printFormatDictionary.get(name));
		return f;
	}
	/** Get a list of the print formats. */
	public String[] getPrintFormats() {
		return getDictKeys(printFormatDictionary);
	}

	/** Add one record format to our dictionary */
	public void addRecordFormat(RecordFormat f) {
		recordFormatDictionary.put(f.getName(),f);
	}
	/** Get a record format by name. */
	public RecordFormat findRecordFormat(String name) {
		RecordFormat f =
			(RecordFormat)(recordFormatDictionary.get(name));
		return f;
	}
	/** Get a list of the record formats. */
	public String[] getRecordFormats() {
		return getDictKeys(recordFormatDictionary);
	}

	public void setInputStream(InputStream rawIn) {
		inputStream= new LineNumberReader(new InputStreamReader(rawIn));
	}
	public void setRootsBase(RootsBase b) {
		base = b;
		base.ci = this;
	}

	/** Run the CommandInterpreter thread when caller invokes 'start' */
	public void run() {
		runMe = true;
		if (inputStream==null)
			setInputStream(System.in);
		if (base==null)
			setRootsBase(new RootsBase());
		while (runMe) {
			/* Set prompt to null to skip printing it, such as
			 * for batch-mode input. */
			if (prompt!=null) {
				/* Print prompt if there */
				System.out.print(prompt);
				System.out.flush();
			}
			try {
				String line = inputStream.readLine();
				if (line==null)
					break;	/* EOF */
//System.out.println("Command line: "+line);
				doCommand(line);
			} catch (Exception e) {
				System.out.println("Exception:"+e.getMessage());
				e.printStackTrace();
			}
		}
		/* Done running */
	}
	/** Parse and execute a single command line */
	public void doCommand(String line) {
		String words[] = parseCommandLine(line);
		doCommand(words);
	}

	/** Parse a single command line into an array of strings. */
	public String[] parseCommandLine(String line) {
		return StringParser.parse(line);
	}

	/** Execute a single command from an array of word strings */
	public void doCommand(String words[]) {
		if (words.length==0)
			return;		/* empty command */
		Command c = findCommand(words[0]);
		if (c!=null) {
			c.action(words);
		} else {
			System.out.println("Unknown command '"+words[0]+"'");
			/* TBD - give warning about unknown command */
		}
	}

	/** Get a list of the keys from a dictionary. */
	private String[] getDictKeys(Dictionary d) {
		String sa[] = new String[d.size()];
		Enumeration e = d.keys();
		int n = 0;
		while (e.hasMoreElements())
			sa[n++] = (String)e.nextElement();
		Sort.sortStrings(sa);	/* sort in alpha order */
		return sa;
	}
}

/* end */
