/* PolyModel.java
 *
 * Jim McBeath, May 24, 2002
 * Merged from ListModel and SingleModel of Jun 7, 2002
 */

package jimmc.roots;

import jimmc.treepar.Box;
import jimmc.treepar.Connector;
import jimmc.treepar.Layout;
import jimmc.util.ArrayUtil;
import jimmc.util.StringUtil;

import java.io.ByteArrayOutputStream;
import java.io.PrintWriter;
import java.util.Vector;

/** A model for either a single record or multiple records from the database.
 */
public class PolyModel {
	/** Our database model. */
	protected DatabaseModel db;

	/** Our key list. */
	protected String[] keys;

	/** Our base database. */
	protected RootsBase base;

	/** If only one key, the record for that key. */
	protected RootsRecord singleRecord;

	/** Create a new model. */
	public PolyModel(DatabaseModel db, String[] keys) {
		this.db = db;
		this.keys = keys;
		base = db.getBase();
		if (keys!=null && keys.length==1)
			singleRecord = base.findRecord(keys[0]);
	}

	/** Get a selection relative to this selection.
	 * @param what the selction, such as father, mother, or children.
	 */
	public String[] getSelection(String what) {
		db.select(what+" of "+StringUtil.toString(keys,' '));
		return db.getSelectedKeys();
	}

	/** Get the keys for this model. */
	public String[] getKeys() {
		return keys;
	}

	/** Get the count of keys. */
	public int getCount() {
		return keys.length;
	}

	/** Get the type for this model. */
	public String getType() {
		if (keys==null)
			return "none";
		if (keys.length>1)
			return "list";
		if (singleRecord==null)
			return "(none)";
		return singleRecord.formatNcvType();	/* TBD - do better */
	}

	/** Get the summary line to display for this model. */
	public String getSummary() {
		if (singleRecord==null)
			return "Multiple records: "+getCount();
		return singleRecord.formatOneLine(new PrintOptions());
	}

	/** Get the command we use to generate our display. */
	public String getDisplayCommand(String format, boolean includeKeys) {
		StringBuffer sb = new StringBuffer();
		sb.append("print ");
		if (includeKeys)
			sb.append("-k ");
		if (format.equals("list"))
			sb.append("-f list -r oneLine");
		else if (format.equals("gedcom"))
			sb.append("-f list -r gedcom");
		else if (format.equals("ncv"))
			sb.append("-r ncv");
		else {
			sb.append("-f ").append(format).append(" ");
			if (format.equals("atree")) {
				if (keys.length==1)
					sb.append("plus recurse parents of ");
				//If multiple keys, use that list
			}
			else if (format.equals("dtree")) {
				if (keys.length==1)
					sb.append("plus recurse children of ");
				//If multiple keys, use that list
			}
		}
		for (int i=0; i<keys.length; i++)
			sb.append(' ').append(keys[i]);   /* our records */
		return sb.toString();
	}

	/** Get the long text string based on format. */
	public String getOutputText(String format, boolean includeKeys) {
		if (keys==null || keys.length==0)
			return "";
		if (format.equals("Ttree")) {
			//Ttree is a text version of the Gtree format,
			//for debugging.
			return getTreeparText(format,includeKeys);
		}

		String cmd = getDisplayCommand(format,includeKeys);
		db.logCommand(cmd);
		return db.doCommand(cmd);
	}

	/** Get the text dump of a treepar layout. */
	protected String getTreeparText(String format, boolean includeKeys) {
		Layout layout = getOutputLayout(format, includeKeys);

		/* Create a string output stream to write to. */
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		PrintWriter ps = new PrintWriter(baos);
		layout.dump(ps);	/* dump to the string */
		return baos.toString();	/* return the dump */
	}

	/** Get the layout and route it. */
	protected Layout getOutputLayout(String format, boolean includeKeys) {
		Layout layout = new Layout(true);

		PrintOptions po = new PrintOptions();
		po.includeKeys = includeKeys;

		String[] sel = keys;	//default selection is our keys
		if (format.equals("Gtree")||format.equals("Ttree")) {
			if (keys.length==1) {
				String key = keys[0];
				String[] refKeys =
					singleRecord.getCtreeReferenceKeys();
				if (refKeys!=null)
					sel = ArrayUtil.union(keys,refKeys);
			}
		} else {
			throw new RuntimeException(
				"Unknown layout format "+format);
		}

		sel = pruneRedundantTreeKeys(sel);

		for (int i=0; i<sel.length; i++) {
			String key = sel[i];
			RootsRecord record = base.findRecord(key);
			record.layoutBox(layout,po,(i==0)?key:null);
		}

		/* TBD - add layout.setRootBox(firstBox);
		 * add method to Layout.java */
		//layout.placeAndRoute(null,Layout.STEP_ROW_NUMBERS);
		layout.placeAndRoute();
		return layout;
	}

	/** Prune the list of items to remove redundant nodes.
	 * RootsPerson nodes are replaced by their parent union node
	 * (if defined), then duplicate nodes are removed from the list.
	 */
	protected String[] pruneRedundantTreeKeys(String[] sel) {
		Vector v = new Vector();
		for (int i=0; i<sel.length; i++) {
			String key = sel[i];
			RootsRecord record = base.findRecord(key);
			if (record instanceof RootsPerson) {
				RootsPerson p = (RootsPerson)record;
				RootsUnion u = p.findParentUnion();
				if (u!=null) {
					record = u;	//switch to parents
					key = record.getKey();
				}
			}
			if (!v.contains(key))
				v.addElement(key);
		}
		String[] kk = new String[v.size()];
		v.copyInto(kk);
		return kk;
	}

	/** Debug - Fill in the layout with boxes. */
	protected void fillLayout(Layout layout) {
		Box box;
		Connector conn;

		box = new Box(layout,"b1");
		box.setSize(40,60);
		box.setText("This is b1");
		conn = new Connector(box,"net1");
		conn.setSide(Layout.EAST);
		conn.setPosition(40,30);

		box = new Box(layout,"b2");
		box.setSize(45,50);
		box.setText("Box b2");
		conn = new Connector(box,"net1");
		conn.setSide(Layout.WEST);
		conn.setPosition(0,20);
	}
}

/* end */
