/* Copyright 1997 Jim McBeath
 *
 * Jim McBeath, January 12, 1997
 */

package jimmc.roots;

import jimmc.util.ResourceSource;

import java.io.*;
import java.text.MessageFormat;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.MissingResourceException;
import java.util.ResourceBundle;
import java.util.StringTokenizer;
import java.util.Vector;

/** RootsBase is the database object for a simple in-memory genealogy
 * database.
 */

public class RootsBase implements ResourceSource {
	CommandInterpreter ci;
	RootsRecord data[];
	private Vector vData;	/* used while reading data file */
	private RootsRecord[] selection;
	private Hashtable keyDict;

	public PrintOptions defaultPrintOptions =
		new PrintOptions(new PrintListBlank(ci),new FormatNcv(ci));

	public PrintOptions defaultShowOptions =
		new PrintOptions(new PrintList(ci), new FormatKeyLine(ci));

	public PrintOptions currentPrintOptions = defaultPrintOptions;

	protected ResourceBundle resources;

	/** Create an empty roots database */
	public RootsBase() {
		initResources();
	}

	/** Get the number of records in the database. */
	public int getRecordCount() {
		if (data==null)
			return 0;
		return data.length;
	}

	/** Read jimmc's name:value database file into the database.
	 * Ncv means name-colon-value, such as "SOURCE: John Doe" */
	public void readNcvFile(String filename) {
		FileReader inFile;
		LineNumberReader li;
		try {
			inFile = new FileReader(filename);
			li = new LineNumberReader(inFile);
		} catch (IOException e) {
			println("Error opening file "+filename);
			/* TBD */
			return;
		}
		/* Read each record from inFile, add to our data */
		vData = new Vector();
		keyDict = new Hashtable();
		try {
			while (readNcvRecord(li)) ;
				/* read every record */
		} catch (Exception e) {
			throw new RuntimeException(
				"Exception at line "+li.getLineNumber()+
					": "+e.getMessage());
					/* TBD */
		}
		try {
			inFile.close();	/* done with file */
		} catch (Exception e) {
			/* TBD - handle IOException */
		}
		data = new RootsRecord[vData.size()];
		vData.copyInto(data);
		println(data.length+" records loaded");
		/* Fix up the links between records */
		for (int i=0; i<data.length; i++) {
			data[i].link();
			data[i].parseDates();
		}
		if (data.length>0) {
			selection = data[0].asArray();
		} else
			selection = null;
	}
	/* Read one record from the input file, parse it and insert into DB.
	 * @return true to keep reading, false if done */
	private boolean readNcvRecord(LineNumberReader li) throws IOException {
		RootsRecord r;
		int lineNumber = li.getLineNumber();
		String firstLine = li.readLine();  /* get key line */
		while (firstLine!=null && firstLine.equals(""))
			firstLine = li.readLine(); /* skip blank lines */
		if (firstLine==null)
			return false;	/* no more data */
		String key = Ncv.getKey(firstLine);
		String lines[] = Ncv.readLines(li);
			/* read in one record of lines from the file */
		if (key==null) {
			println("Null key at line "+lineNumber);
			return true;
		}
		if (key.equals("0"))
			return true;	/* ignore this record */
		if (lines.length==0)
			return true;	/* empty record */
		String recordType = Ncv.getRecordValue(lines,"T");
		if (recordType==null)
			return true;	/* ignore this record */
		if (recordType.equals("I"))
			r = RootsPerson.ParseNcvLines(this,lines);
		else if (recordType.equals("F"))	/* Family = Union */
			r = RootsUnion.ParseNcvLines(this,lines);
		else if (recordType.equals("A"))
			r = RootsAddress.ParseNcvLines(this,lines);
		else if (recordType.equals("R"))	/* Reference = Source */
			r = RootsSource.ParseNcvLines(this,lines);
		else {
			println("Line "+lineNumber+": "+
			    "Unknown input record type 'T:"+recordType+"'");
		/* TBD - print warning message with line number */
			return true;	/* ignore and keep reading */
		}
		r.key = key;
		r.keyComment = firstLine.substring(key.length());
		vData.addElement(r);
		keyDict.put(key,r);
		return true;	/* keep reading */
	}

	/** Given a key value, return the record for that key */
	public RootsRecord findRecord(String key) {
		if (key==null)
			return null;
		RootsRecord r = (RootsRecord)(keyDict.get(key));
		return r;
	}

	/** Given a comma-separated list of key values, return an
	 * array of records for those keys.  If any items in the list
	 * are not found, the corresponding entry in the array is null. */
	public RootsRecord[] findRecordList(String keylist) {
		if (keylist==null || keylist.equals(""))
			return null;
		StringTokenizer st = new StringTokenizer(keylist,", ");
		int c = st.countTokens();
		if (c==0)
			return null;
		RootsRecord ra[] = new RootsRecord[c];
		c = 0;
		while (st.hasMoreTokens()) {
			ra[c++] = findRecord(st.nextToken());
		}
		return ra;
	}

	/** Print the current selection in the specified format. */
	public void printRecords(PrintOptions po) {
		if (selection==null) {
			println("No current selection");
			return;
		}
		printRecords(po,selection);
	}
	/** Print the specified selection using the given print options.
	 */
	public void printRecords(PrintOptions po, RootsRecord ra[]) {
		PrintOptions oldPrintOptions = currentPrintOptions;
		try {
			currentPrintOptions = po;
			po.printRecords(ra);
		} finally {
			currentPrintOptions = oldPrintOptions;
		}
	}

	/** Print out the data for the given selection expression */
	public void printRecords(PrintOptions po,
				String what[],int start,int count) {
		if (count==0) {
			printRecords(po);
			return;
		}
		RootsRecord ra[] = findSelection(what,start,count);
		if (ra==null) {
			println("No match");
			return;
		}
		printRecords(po,ra);
	}

	/** Clear the current selection.
	 */
	public void clearSelection() {
		selection = null;
	}

	/** Change the selection to the referenced selection.
	 * @param what array of strings containing selection expression
	 * @param start index of first item in selection expression
	 * @param count number of items in selection expression
	 */
	public void select(String what[], int start, int count) {
		RootsRecord ra[];
		ra = findSelection(what,start,count);
		if (ra==null) {
			println("No match");
			return;
		}
		selection = ra;
		printRecords(defaultShowOptions,ra);
	}

	/** Get the records referred to by the specified selection expression.
	 * @param what array of selection strings in expression
	 * @param start index of first item to consider
	 * @param count number of items to consider
	 * @return array of matching records if found, null if not found
	 */
	public RootsRecord[] findSelection(String what[],int start,int count) {
		return findSelection(what,start,count,selection);
	}
	/** Get a selection by expression relative to another selection.
	 * @param what array of selection strings in expression
	 * @param start index of first item to consider
	 * @param count number of items to consider
	 * @param selection the selection this expression is relative to
	 * @return array of matching records if found, null if not found
	 */
	public RootsRecord[] findSelection(String what[], int start, int count,
					RootsRecord selection[]) {
		if (count==1) {
			return findSelection(what[start],selection);
		}

		/* Look for "something of someone" */
		for (int c=count-2; c>0; c--) {
		    if (what[start+c].equals("of")) {
			/* Selection is in the form "something of someone",
			 * e.g. "wife of 3001", so we first figure out the
			 * "of someone" part, then use that as the selection
			 * to figure out the "something of" part. */
			RootsRecord ra[] = findSelection(what,
				start+c+1, count-c-1, selection);
			if (ra==null)
				return null;
			return findSelection(what,start,c,ra);
		    }
		}

		/* No "of" in this selection, but it's more than one word,
		 * look for special cases. */

		/* Process "recurse" keyword. */
		if (what[start].startsWith("recurse"))
			return findSelectionRecurse(what,start,count,selection);

		/* Process "plus" keyword. */
		if (what[start].equals("plus"))
			return findSelectionPlus(what,
				start+1,count-1,selection);

		/* Process name matching operators. */
		if (what[start].equals("name"))
			return findSelectionField(what,
				start+1,count-1,selection,"name");

		/* Process place matching operators. */
		if (what[start].equals("place"))
			return findSelectionField(what,
				start+1,count-1,selection,"place");

		/* Process key matching operators. */
		if (what[start].equals("key"))
			return findSelectionField(what,
				start+1,count-1,selection,"key");

		/* Go through each one and put them all
		 * together.  This allows selections such as "id1 id2"
		 * and "child0 child3 wife" */
		RootsVector vr = new RootsVector();
		for (int c=0; c<count; c++) {
			RootsRecord ra[];
			ra = findSelection(what[start+c],selection);
			vr.add(ra);
		}
		return vr.toRootsRecordArray();
	}

	/** Process the "recurse" keyword in a selection.
	 * The recurse selection says we should make the
	 * initial reference on the given selection, then
	 * keep on making that reference on the returned
	 * selection until we don't have any more new items.
	 * For example, "recurse father" will get all of the
	 * male ancestors. */
	protected RootsRecord[] findSelectionRecurse(String[] what, int start,
			int count, RootsRecord[] selection) {
		/* Look for recursion limit, if there. */
		String numstr = what[start].substring(7);
		RootsRecord ra[] = findSelection(what,
			start+1, count-1, selection);
				/* do the reference the first time */
		if (ra==null)
			return null;
		int num;
		if (numstr.equals(""))
			num = -1;
		else
			num = Integer.parseInt(numstr);
		RootsVector vr = new RootsVector();
		vr.add(ra);

		/* Check for limited recursion.
		 * recurse0 is same as recurse; no limit.
		 * recurse1 is same as no recurse; we already
		 *  did the reference once above, so we don't
		 *  do any more in this loop.
		 * recurse2 says do the reference twice, i.e.
		 *  once just above, and once in this loop.
		 * etc. for higher recursion numbers.
		 */
		RootsRecord r1[];
		int c = 0;
		while (--num != 0) {
			RootsVector vn = new RootsVector();
			while (c<vr.size()) {
				r1 = vr.recordAt(c).asArray();
				ra = findSelection(what,start+1,
					count-1,r1);
				vn.add(ra);
				c++;
			}
			vr.add(vn.toRootsRecordArray());
			if (c>=vr.size())
				break;	/* no more */
		}
		return vr.toRootsRecordArray();
	}

	/** Process the "plus" keyword in a selection.
	 * The plus keyword says to take the following selection
	 * and add it to its source.  For example, the selection
	 * "plus wife of father of 123" will select the father
	 * and the wife. */
	protected RootsRecord[] findSelectionPlus(
				String[] what, int start,
				int count, RootsRecord[] selection) {
		RootsRecord ra[];
		ra = findSelection(what,start,count,selection);
		if (selection==null)
			return ra;
		if (ra==null)
			return selection;
		RootsVector vr = new RootsVector();
		vr.add(selection);	/* old selection first */
		vr.add(ra);		/* then new selection */
		return vr.toRootsRecordArray();
	}

	/** Process the name matching operations. */
	protected RootsRecord[] findSelectionField(
				String[] what, int start,
				int count, RootsRecord[] selection,
				String fieldName) {
		int op;
		String pattern = what[start+1];
		if (what[start].equals("includes"))
			pattern = "*"+pattern+"*";
		else if (what[start].equalsIgnoreCase("startswith"))
			pattern = pattern+"*";
		else if (what[start].equals("matches"))
			; /* leave as is */
		else {
			println("Unknown match operator '"+what[start]+"'");
			return null;
		}

		/* Look through all records for matching field. */
		Enumeration enum = keyDict.keys();
		RootsVector vr = new RootsVector();
		while (enum.hasMoreElements()) {
			String key = (String)enum.nextElement();
			RootsRecord r = (RootsRecord)(keyDict.get(key));
			if (r.fieldMatches(fieldName,pattern))
				vr.add(r);
		}
		return vr.toRootsRecordArray();
	}

	/** Get a selection by simple expression relative to another selection.
	 * @param what one selection string.
	 * @return array of matching records if found, null if not found.
	 */
	public RootsRecord[] findSelection(String what,RootsRecord selection[]){
		RootsRecord r;
		RootsRecord ra[];
		r = findRecord(what);	/* see if it's a key */
		if (r!=null)
			return r.asArray();	/* found it by key */
		if (selection==null)
			return null;
		if (what.startsWith("current"))
			return selection[0].findListReference(
				selection,what.substring(7));
		if (what.equals("all"))
			return data;	/* the whole list */
		RootsVector vr = new RootsVector();
		for (int i=0; i<selection.length; i++) {
			ra = selection[i].findReference(what);
			vr.add(ra);
		}
		return vr.toRootsRecordArray();
	}

	/** Get the list of currently selected keys. */
	public String[] getSelectedKeys() {
		if (selection==null)
			return new String[0];	/* no keys */
		String[] keys = new String[selection.length];
		for (int i=0; i<selection.length; i++)
			keys[i] = selection[i].key;
		return keys;
	}

	/** Get the list of currently selected keys as a single string. */
	public String getSelectedKeysString() {
		String[] keys = getSelectedKeys();
		StringBuffer sb = new StringBuffer();
		for (int i=0; i<keys.length; i++) {
			if (i>0)
				sb.append(" ");
			sb.append(keys[i]);
		}
		return sb.toString();
	}

	/** Print a line to our current output */
	public void println(String s) {
		currentPrintOptions.out.println(s);
		currentPrintOptions.out.flush();
	}

	/** Set up our resources. */
	public void initResources() {
		resources = ResourceBundle.getBundle(
			"jimmc.roots.Resources");
	}

	/** Get a string from our resource file. */
	public String getResourceString(String name) {
		try {
			return resources.getString(name);
		} catch (MissingResourceException ex) {
			return name;	//use the resource name
		}
	}

	/** Get a string from a resource file and use it as the format in
	 * a call to MessageFormat.format.
	 * args.
	 * @param name The name of the resource.
	 * @param args The args to pass to MessageFormat.format.
	 * @return The formatted string.
	 */
	public String getResourceFormatted(String name, Object[] args){
		String fmt = getResourceString(name);
		return MessageFormat.format(fmt,args);
	}

	/** Get a string from a resource file and format it with the given
	 * arg.
	 */
	public String getResourceFormatted(String name, Object arg) {
		String fmt = getResourceString(name);
		Object[] args = { arg };
		return MessageFormat.format(fmt,args);
	}

	/** Sample main program */
	public static void main(String args[]) {
		RootsBase b = new RootsBase();
			/* Initialize an empty database.  Let the user
			 * fill it in with his data file using the load
			 * command in the CommandInterpreter. */
		CommandInterpreter c = new CommandInterpreter(b);
		Thread commandThread =
			new Thread(c,"RootsBaseCommandInterpreter");
		commandThread.start();	/* start the command interpreter */
		/* We're done; we return from this main method with the
		 * command interpreter still running.  When it finishes,
		 * all threads will be done and java will exit.
		 */
	}
}

/* end */
