/* RootsBox.java
 *
 * Jim McBeath, July 16, 1998
 */

package jimmc.roots;

import jimmc.treepar.Box;
import jimmc.treepar.Connector;
import jimmc.treepar.Layout;

/** RootsBox is a version of Box specialized for use with the roots package.
 */
public class RootsBox extends Box {
	/** Our text. */
	protected StringBuffer textBuffer;

	/** The key of our father. */
	protected String fatherKey;

	/** The key of our mother. */
	protected String motherKey;

	/** The current connector location.
	 * Starts at 0 and goes more negative with each connector.
	 * When done() is called, all connector positions are adjusted.
	 */
	protected int currentY;

	/** The offset of the parent connectors from the top/bottom */
	public int parentConnOffset;

	/** Create a new box. */
	public RootsBox(Layout layout, String name) {
		super(layout, name);
		textBuffer = new StringBuffer();
		currentY = -layout.getLineHeight();
		setTextPosition(layout.getCharacterWidth(),
				layout.getLineHeight()/2);
		parentConnOffset = layout.getLineHeight();
	}

	/** Add a section of text, optionally with a child connector.
	 * @param text The text for this section.  This text gets appended
	 *        to the text in the Box.
	 * @param childKey The key of the child for the east-side connector
	 *        for this section, or null if no connector.
	 */
	public void addSection(String text, String childKey) {
		if (!text.endsWith("\n"))
			text = text+"\n";
		textBuffer.append(text);

		if (childKey!=null)
			addConnector(text,childKey);

		int numLines = countLines(text);
		currentY -= numLines * layout.getLineHeight();
	}

	/** Add a connector. */
	protected void addConnector(String text, String childKey) {
		int firstLength = firstLineLength(text);
		Connector conn = new Connector(this,childKey);
		conn.setPosition((firstLength+1) * layout.getCharacterWidth(),
					currentY);
		conn.setSide(Layout.EAST);
	}

	/** Add parent connectors.
	 * @param fatherKey Key for the father, used for a west connector
	 *        near the top of the box.  Null means no connector.
	 * @param motherKey Key for the mother, used for a west connector
	 *        near the bottom of the box.
	 */
	public void setParentKeys(String fatherKey, String motherKey) {
		this.fatherKey = fatherKey;
		this.motherKey = motherKey;
	}

	/* Everything is done, set up the Box.
	 */
	public void done() {
		String text = textBuffer.toString();
		setText(text);

		int height = (countLines(text)+1) * layout.getLineHeight();
		int width = countWidth(text) * layout.getCharacterWidth();
		setSize(width,height);

		setupDone();	/* set up connectors array */

		/* now add our height to all connectors' Y positions */
		for (int i=0; i<connectors.length; i++)
			connectors[i].movePosition(0,height);

		/* Set up parent connectors */
		if (fatherKey!=null) {
			Connector conn = new Connector(this,fatherKey);
			conn.setPosition(0,height-parentConnOffset);
			conn.setSide(Layout.WEST);
		}
		if (motherKey!=null) {
			Connector conn = new Connector(this,motherKey);
			conn.setPosition(0,parentConnOffset);
			conn.setSide(Layout.WEST);
		}
	}

	/** Count the number of newlines found in the string. */
	protected int countLines(String s) {
		int x = 0;
		int n = 0;
		while ((x=s.indexOf('\n',x))>=0) {
			n++;
			x++;
		}
		return n;
	}

	/** Count the length of the longest line. */
	protected int countWidth(String s) {
		int x = 0;
		int n = 0;
		int xx;
		while ((xx=s.indexOf('\n',x))>=0) {
			int nn = xx - x;	/* length of this line */
			if (nn>n)
				n = nn;		/* keep the max */
			x = xx+1;
		}
		return n;
	}

	/** Return the number of characters in the first line. */
	protected int firstLineLength(String s) {
		return s.indexOf('\n');		/* assume str ends with LF */
	}
}

/* end */
