/* TextPopup.java
 *
 * Jim McBeath
 */

package jimmc.roots;

import java.awt.MenuItem;
import java.awt.PopupMenu;

/** A popup menu for use in Text areas.
 * It includes a select command that picks the key value (id field) out
 * of the text under the cursor.
 */
public class TextPopup extends PopupMenu {
	TextPopupListener w;
	MenuPush selectMenu;
	String id;

	/** Create the popup. */
	public TextPopup(TextPopupListener ww) {
		super("------");
		w = ww;
		addMenuItems();
	}

	/** Add the menu items to the popup. */
	protected void addMenuItems() {
		add(selectMenu=new MenuPush("Select") {
			public void action() {
				w.processSelect(id);
			}
		});
		add(selectMenu=new MenuPush("Remove From Selection") {
			public void action() {
				w.processRemove(id);
			}
		});
	}

	/** Enable or disable the menus that depend on a valid ID. */
	protected void enableIdMenus(boolean enable) {
		if (selectMenu!=null)
			selectMenu.setEnabled(enable);
	}

	/** Set stuff in the popup menu based on x,y position */
	protected void setRecord(TextAreaWithPopup tf, int x,int y) {
		id = getID(tf,x,y);
		if (id==null) {
			enableIdMenus(false);
			id = "????";
		} else {
			enableIdMenus(true);
		}
		setLabel(id);
	}

	/** Get the ID (key) value from the text under the cursor. */
	private String getID(TextAreaWithPopup tf, int x, int y) {
		int pos = tf.yPixelToPosition(y);
		String text = tf.getText();
		int ePos = text.indexOf('\n',pos);
		if (ePos<0)
			ePos = text.length();
		/* The line of text is between pos and ePos.
		 * Now we look for an ID.
		 */
		int bPos = text.indexOf('[',pos);
		if (bPos>=0 && bPos<ePos) {
			/* found an open square bracket, look for end */
			int bePos = text.indexOf(']',bPos+1);
			if (bePos>=0 && bePos<ePos) {
				/* found open and closed, assume the text
				 * between is the ID string. */
				return text.substring(bPos+1,bePos);
			}
		}
		/* Didn't find matching square brackets, look at start of line*/
		int sPos = text.indexOf(' ',pos);
		int s2Pos = text.indexOf('\t',pos);
		if (s2Pos>=0 && s2Pos<sPos)
			sPos = s2Pos;
		if (sPos>pos)
			return text.substring(pos,sPos);
				/* Well, give that string a try */
		return null;	/* can't figure it out, give up */
	}

	/** Set stuff in the popup menu based on x,y position */
	protected void setRecord(JTextAreaWithPopup tf, int x,int y) {
		id = getID(tf,x,y);
		if (id==null) {
			enableIdMenus(false);
			id = "????";
		} else {
			enableIdMenus(true);
		}
		setLabel(id);
	}

	/** Get the ID (key) value from the text under the cursor. */
	private String getID(JTextAreaWithPopup tf, int x, int y) {
		int pos = tf.xyPixelToPosition(x,y);
		String text = tf.getText();
		int ePos = text.indexOf('\n',pos);
		if (ePos<0)
			ePos = text.length();
		int bPos = pos;
		if (bPos>0)
			bPos = text.lastIndexOf('\n',pos);
		if (bPos<0)
			bPos = 0;
		else
			bPos++;	/* start after the previous newline */
		/* bPos and ePos are the start and end of the line of text
		 * containing the current point.
		 */
		String line = text.substring(bPos,ePos);
		pos -= bPos;

		/* We now have the line of text in 'line', and the position
		 * of the pointer relative to that line in 'pos'.
		 */
		String word = getWordAround(line,pos);
		return word;	/* may be null */
	}

	/** Get the word in the given single line 
	 * around the specified position. */
	protected String getWordAround(String text,int pos) {
		/* Constrain the text to be between spaces */
		int bPos = text.lastIndexOf(' ',pos);
		if (bPos<0)
			bPos = 0;
		int ePos = text.indexOf(' ',pos);
		if (ePos<0)
			ePos = text.length();

		/* Look for enclosing square brackets. */
		int pos0 = text.lastIndexOf('[',pos);
		int pos1 = text.indexOf(']',pos);
		if (pos0>=bPos && pos1>=0 && pos1<=ePos) {
			/* found enclosing square brackets,
			 * between is the ID string. */
			return text.substring(pos0+1,pos1);
		}

		/* Look for enclosing '@' chars (as in gedcom). */
		pos0 = text.lastIndexOf('@',pos);
		pos1 = text.indexOf('@',pos);
		if (pos0>=bPos && pos1>=0 && pos1<=ePos) {
			/* found enclosing "@" chars,
			 * between is the ID string. */
			return text.substring(pos0+1,pos1);
		}

		/* Didn't find matching square brackets, look at start of line*/
		int sPos = text.indexOf(' ');
		int s2Pos = text.indexOf('\t');
		if (s2Pos>=0 && s2Pos<sPos)
			sPos = s2Pos;
		if (sPos>pos) {
			return text.substring(0,sPos);
				/* Well, give that string a try */
		}

		return null;	/* can't figure it out, give up */
	}
}

/* end */
