/* HtmlPanel.java
 *
 * Jim McBeath, October 12, 2001
 */

package jimmc.swing;

import jimmc.util.LimitedList;
import jimmc.util.ResourceSource;

import java.awt.Color;
import java.io.IOException;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;
import javax.swing.JComponent;
import javax.swing.JEditorPane;
import javax.swing.JScrollPane;
import javax.swing.text.Document;
import javax.swing.text.html.HTMLDocument;
import javax.swing.text.html.HTMLEditorKit;

/** A panel to display HTML text. */
public class HtmlPanel extends JScrollPane implements HyperlinkListener {
	/** Initial history size. */
	public static final int INITIAL_HISTORY_SIZE = 100;

	/** Our resource source. */
	protected ResourceSource res;

	/** The editor panel. */
	protected JEditorPane editor;

	/** Our history. */
	protected LimitedList history;

	/** The index of the currently displayed item. */
	protected int historyPosition;

	/** The location of the currently displayed page, or null
	 * if we are displaying an internal string. */
	protected String currentUrl;

	/** Create the Html panel. */
	public HtmlPanel(ResourceSource res) {
		this.res = res;
		editor = new JEditorPane();
		setViewportView(editor);
		editor.setContentType("text/html");
		editor.setEditable(false);
		editor.addHyperlinkListener(this);
		history = new LimitedList(INITIAL_HISTORY_SIZE);
		historyPosition = -1;
	}

	/** Get the component to use for printing. */
	public JComponent getPrintableComponent() {
		return editor;
	}

	/** Set the background color. */
	public void setBackground(Color color) {
		super.setBackground(color);
		if (editor!=null)
			editor.setBackground(color);
	}

	/** Get the location of the currently displayed page, or null if
	 * the page is internally generated (was set by a call to
	 * {@link #showHtml}).
	 */
	public String getUrl() {
		return currentUrl;
	}

	/** Show the contents of a URL. */
	public void showUrl(String url) {
		showUrlNoHistory(url);
		addToHistory(url);
	}

	/** Show the contents of a URL, but don't add it to the history.
	 */
	public void showUrlNoHistory(String url) {
		if (url.startsWith("html:")) {
			showHtmlNoHistory(url.substring(5));
			currentUrl = url;
			return;
		}
		Document doc = editor.getDocument();
		try {
			editor.setPage(url);
		} catch (IOException ex) {
			editor.setDocument(doc);	//reset to the old page
			String msg = res.getResourceFormatted(
				"error.CantOpenUrl",url);
			showHtmlNoHistory(msg);
		}
		currentUrl = url;
	}

	/** Show an html string. */
	public void showHtml(String html) {
		showHtmlNoHistory(html);
		addToHistory("html:"+html);
	}

	/** Show an html string, but don't add it to the history.
	 */
	public void showHtmlNoHistory(String html) {
		//When we call editor.setText after editor.setPage, and then
		//go back and call setPage on the same URL, it doesn't show
		//us the URL. Apparently it thinks it is still showing that
		//URL.  So we clear out the document here to make it know.
		editor.setEditorKit(new HTMLEditorKit());
		editor.setContentType("text/html");
		editor.setText(html);
		currentUrl = null;
	}

	/** Add an item to the history list. */
	public void addToHistory(String item) {
		if (historyPosition<history.size()-1) {
			//Discard all items past the current position
			history.truncate(historyPosition+1);
		}
		history.addLimited(item);
		historyPosition = history.size()-1;
		fireHistoryChangeEvent();
	}

	/** Back up the position by one and show the resulting item.
	 * @return True if we were able to back up; false if we are
	 *         already at the beginning of the list.
	 */
	public boolean showBack() {
		if (!hasBack())
			return false;	//already at the start of the list
		historyPosition--;
		showUrlNoHistory((String)(history.get(historyPosition)));
		fireHistoryChangeEvent();	//position changed
		return true;
	}

	/** Move the position forward by one and show the resulting item.
	 * @return True if we were able to go forward; false if we are
	 *         already at the end of the list.
	 */
	public boolean showForward() {
		if (!hasForward())
			return false;	//already at the end of the list
		historyPosition++;
		showUrlNoHistory((String)(history.get(historyPosition)));
		fireHistoryChangeEvent();	//position changed
		return true;
	}

	/** True if a call to showBack will work. */
	public boolean hasBack() {
		return (historyPosition>0);
	}

	/** True if a call to showForward will work. */
	public boolean hasForward() {
		return (historyPosition<history.size()-1);
	}

	/** Here when the history changes.
	 * Eventually, this method could have a listener list that it invokes,
	 * but for now we just let users subclass this class and override
	 * this method.
	 */
	public void fireHistoryChangeEvent() {
		//do nothing
	}

    //The HyperlinkListener interface
	/** Handle a hyperlink event. */
	public void hyperlinkUpdate(HyperlinkEvent ev) {
		if (ev.getEventType()==HyperlinkEvent.EventType.ACTIVATED) {
			showUrl(ev.getURL().toString());
		}
	}
    //End HyperlinkListener interface
}

/* end */
