/* Box.java
 *
 * Jim McBeath, July 8, 1998
 */

package jimmc.treepar;

import java.awt.Dimension;
import java.awt.Point;
import java.io.PrintWriter;
import java.util.Vector;

/** Box is the primary visual object in a Layout.
 * The following fields
 * in the Box object must be set: name, size, text, textPosition.
 * @see Layout
 */
public class Box {
    /* These fields are used as input to the place-and-route process. */

	/** The layout we are a part of. */
	protected Layout layout;

	/** The name of this box. */
	protected String name;

	/** The size of the box. */
	protected Dimension size;

	/** The text to display inside the box. */
	protected String text;

	/** The position of the text within the box. */
	protected Point textPosition;

	/** Our connectors. */
	protected Connector[] connectors;

    /* These fields are used during setup phase. */

	/* Our connectors, during setup. */
	protected Vector connectorVector;

    /* These values are set during the place-and-route process. */

	/** The text split up into single lines. */
	protected String[] textLines;

	/** The row number for this box. */
	protected int rowNumber;

	/** True when rowNumber has been set. */
	protected boolean rowNumberSet;

	/** The row containing this box. */
	protected Row row;

	/** The ordering of this box within its row. */
	protected int rowOrder;

	/** True when the row ordering has been set for this box. */
	protected boolean orderSet;

	/** The location of the lower left corner of the box. */
	protected Point origin;

	/** True after we have fixed the position of the box in the row. */
	protected boolean positionSet;

    /* Methods */

	/** Create a new Box in a layout. */
	public Box(Layout layout, String name) {
		this.layout = layout;
		this.name = name;
		connectorVector = new Vector();
		layout.addBox(this);
		origin = new Point(0,0);
		size = new Dimension(0,0);
		textPosition = new Point(0,0);
	}
	
	/** Set the size of the Box. */
	public void setSize(int width, int height) {
		size = new Dimension(width,height);
	}

	/** Set the text inside the Box. */
	public void setText(String text) {
		this.text = text;
	}

	/** Set the location of the text relative to the box origin. */
	public void setTextPosition(int x, int y) {
		textPosition.x = x;
		textPosition.y = y;
	}

	/** Add a connector to this box.
	 * This method is called from the contructor for Connector.
	 */
	protected void addConnector(Connector connector) {
		connectorVector.addElement(connector);
		connectors = null;
	}

	/** Call this method when setup on this box is done, after all
	 * connectors have been added.
	 */
	public void setupDone() {
		connectorsDone();
		textDone();
	}

	/** Create the array of connectors from the vector. */
	protected void connectorsDone() {
		if (connectors==null) {
			connectors = new Connector[connectorVector.size()];
			connectorVector.copyInto(connectors);
		}
	}

	/** Create the array of text lines from the text string. */
	protected void textDone() {
		if (text==null)
			text = "";
		if (!text.endsWith("\n"))
			text = text+"\n";
		int n = countLines(text);
		textLines = new String[n];
		int x = 0;	/* position of the start of each line */
		for (int i=0; i<n; i++) {
			int e = text.indexOf('\n',x);
			textLines[i] = text.substring(x,e);
			x = e+1;
		}
	}

	/** Get the thickness of this box. */
	protected int getThickness() {
		return (layout.verticalRows)?size.width:size.height;
	}

	/** Get the length of this box. */
	protected int getLength() {
		return (layout.verticalRows)?size.height:size.width;
	}

	/** Clear the row number and pointer for the box. */
	protected void clearRowNumber() {
		rowNumberSet = false;
		rowNumber = 0;
		row = null;
	}

	/** Set the row number for the box. */
	public void setRowNumber(int rowNum) {
		if (rowNumberSet)
			return;		/* already set */
		rowNumber = rowNum;
		rowNumberSet = true;
		/* Propagate through our connectors to other boxes. */
		for (int c=0; c<connectors.length; c++)
			connectors[c].setRowFromBox(rowNum);
	}

	/** Set the row for the box. */
	protected void setRow(Row row) {
		this.row = row;		/* may be null */
	}

	/** Set the R coordinate as specified.
	 */
	protected void setR(int d) {
		if (layout.verticalRows)
			origin.y = d;
		else
			origin.x = d;
	}

	/** Get the R coordinat. */
	protected int getR() {
		return (layout.verticalRows)?origin.y:origin.x;
	}

	/** Change the R coordinate of this box.
	 * @param d the amount to add to the R coordinate.
	 */
	protected void moveR(int d) {
		if (layout.verticalRows)
			origin.y += d;
		else
			origin.x += d;
	}

	/** Set the S coordinate of this box to the same as its row. */
	protected void setS() {
		if (row==null)
			return;
		int S = row.getS();
		if (layout.verticalRows)
			origin.x = S;
		else
			origin.y = S;
	}

	/** Clear the row ordering of this box. */
	protected void clearRowOrder() {
		orderSet = false;
	}

	/** Set the row ordering of this box and neighbors.
	 * @param negdir if set, expand in the negative direction.
	 */
	protected void setRowOrder(boolean negdir) {
		if (orderSet)
			return;	/* already set */
		orderSet = true;
		if (negdir)
			rowOrder = --row.rowOrderMin;
		else
			rowOrder = row.rowOrderMax++;
		/* set the ordering on the connectors */
		setConnectorsRowOrder(layout.NorE,negdir);
		setConnectorsRowOrder(layout.SorW,negdir);
	}

	/** Return true if the row order has already been set on this box.
	 */
	protected boolean isRowOrderSet() {
		return orderSet;
	}

	/** Set the positionSet flag. */
	protected void setPositionSet(boolean b) {
		positionSet = b;
	}

	/** Return true if the positionSet flag has been set. */
	protected boolean isPositionSet() {
		return positionSet;
	}

	/** Set the ordering on the connectors on the specified side
	 * of this box.
	 * @param side the side of the box out of which connectors
	 *        we propagate to the network.
	 * @param negdir true means new items are added to the lesser
	 *        end of the row.
	 */
	protected void setConnectorsRowOrder(int side, boolean negdir) {
		Connector[] cc = Connector.getSideConnectors(connectors,side);
		if (cc==null)
			return;		/* no connectors on that side. */

		/* See if the net on any of these connectors has been set.
		 * If so, then we don't have a simple tree-structured
		 * network.  We will do our continued order around
		 * that connector in an effort to minimize crossed lines.
		 */
		int c;
		for (c=0; c<cc.length; c++)
			if (cc[c].net.isRowOrderSet())
				break;
		if (c>=cc.length) {
			/* no nets have already been set, just start at
			 * one end and work through them.
			 */
			c = negdir?cc.length:0;
		}

		/* Traverse the nets on our connectors that are higher than
		 * the previously found net.
		 */
		for (int j=c; j<cc.length; j++)
			cc[j].net.setRowOrder(false);
		/* Traverse the nets that are lower. */
		for (int j=c-1; j>=0; j--)
			cc[j].net.setRowOrder(true);
	}

	/** Add in the force data for all of the connectors on the
	 * specified side of the box.
	 * @param side ignore connectors not on this side of the box.
	 * @param forcesums [0] is the force sum, [1] is the count.
	 *        We add to these two numbers.
	 */
	protected void addNetForce(int side, int[] forcesums) {
		for (int c=0; c<connectors.length; c++) {
			Connector cc = connectors[c];
			if (cc.getSide()==side)
				cc.net.addNetForce(cc,forcesums);
		}
	}

	/** Increase the bounds if necessary to include this box. */
	protected void addToBounds(BoundsRectangle bounds) {
		bounds.add(origin.x, origin.y);
		bounds.add(origin.x+size.width, origin.y+size.height);
	}

	/** Count the number of newlines found in the string. */
	protected int countLines(String s) {
		int x = 0;
		int n = 0;
		while ((x=s.indexOf('\n',x))>=0) {
			n++;
			x++;
		}
		return n;
	}

	/** Count the length of the longest line.
	 * @return The number of characters in the longest line.
	 */
	protected int countWidth(String s) {
		int x = 0;
		int n = 0;
		int xx;
		while ((xx=s.indexOf('\n',x))>=0) {
			int nn = xx - x;	/* length of this line */
			if (nn>n)
				n = nn;		/* keep the max */
			x = xx+1;
		}
		return n;
	}

	/** Draw the box. */
	protected void paint(LayoutGraphics g) {
		g.drawBox(origin,size);

		/* Figure out the point above the topmost line. */
		Point pText = new Point(origin.x + textPosition.x,
				origin.y + textPosition.y +
					textLines.length * layout.lineHeight);
		/* Draw all the strings, from top to bottom. */
		for (int i=0; i<textLines.length; i++) {
			pText.y -= layout.lineHeight;
			g.drawString(textLines[i],pText);
		}
	}

	/** True if the box contains the specified point. */
	protected boolean contains(int x, int y) {
		if (x<origin.x || y<origin.y)
			return false;
		if (x>origin.x+size.width || y>origin.y+size.height)
			return false;
		return true;
	}

	/** Find the text at the specified point. */
	protected LayoutHitText getHitText(int x, int y) {
		/* Figure out the point above the topmost line. */
		int xMin = origin.x+textPosition.x;
		int yMax = origin.y+textPosition.y +
					textLines.length * layout.lineHeight;
		int yMin = yMax - layout.lineHeight;
		for (int i=0; i<textLines.length; i++) {
			int textLength = textLines[i].length();
			int xMax = xMin + textLength*layout.characterWidth;
				/* TBD - charWidth is not very accurate */
			if (x>=xMin && x<=xMax && y>=yMin && y<=yMax) {
				LayoutHitText hit = new LayoutHitText();
				hit.text = textLines[i];
				hit.position = (x-xMin)/layout.characterWidth;
				return hit;
			}
			yMax -= layout.lineHeight;
			yMin -= layout.lineHeight;
		}
		return null;
	}

	/** Dump a box for debugging. */
	public void dump(PrintWriter f, String prefix, boolean details) {
		f.println(prefix+"Box: "+name);
		if (!details)
			return;
		f.println(prefix+"  rowNumber="+rowNumber+
				", set="+rowNumberSet+
				"; rowOrder="+rowOrder+
				", set="+orderSet);
		f.println(prefix+"  origin="+origin+
				", size="+size);
		Connector.dump(f,connectors,prefix+" ",false);
	}

	/** Dump an array of boxes. */
	public static void dump(PrintWriter f, Box[] boxes, String prefix,
						boolean details) {
		f.print(prefix);
		f.print("Boxes: ");
		if (boxes==null) {
			f.println("null");
			return;
		}
		f.println(boxes.length);
		for (int b=0; b<boxes.length; b++) {
			boxes[b].dump(f,prefix+" ["+b+"] ",details);
		}
	}
}

/* end */
