/* Row.java
 *
 * Jim McBeath, July 8, 1998
 */

package jimmc.treepar;

import java.io.PrintWriter;
import java.util.Vector;

/** Row represents a single row of the layout.
 * There are two types of Row: a box row, and a channel row.
 * A box row contains a set of Box objects; a channel row contains
 * a set of Track objects.
 */
public class Row {
	/** Our layout. */
	protected Layout layout;

	/** Our row number. */
	protected int rowNumber;

	/** Our position (S coordinate). */
	protected int position;

	/** Our thickness. */
	protected int thickness;

	/** Our length. */
	protected int length;

	/** The boxes in this row, during setup. */
	protected Vector boxesVector;

	/** The boxes in the row, after setup. */
	protected Box[] boxes;

	/** The nets in this row, during setup. */
	protected Vector netsVector;

	/** The nets in this row, after setup. */
	protected Net[] nets;

	/** Minimum ordering position of an item in this row. */
	protected int rowOrderMin;

	/** Maximum ordering position of an item in this row. */
	protected int rowOrderMax;

	/** The tracks in this row, when it is a channel row. */
	protected Vector tracksVector;

    /* Methods */

	/** Create a new empty row. */
	public Row(Layout layout, int rowNumber) {
		this.layout = layout;
		this.rowNumber = rowNumber;
		boxesVector = new Vector();
		netsVector = new Vector();
	}

	/** Add a box to this row. */
	protected void addBox(Box box) {
		boxesVector.addElement(box);
		boxes = null;
	}

	/** Add a net to this row. */
	protected void addNet(Net net) {
		netsVector.addElement(net);
		nets = null;
	}

	/** Call this when done adding boxes and nets. */
	protected void setupDone() {
		boxes = new Box[boxesVector.size()];
		boxesVector.copyInto(boxes);
		nets = new Net[netsVector.size()];
		netsVector.copyInto(nets);
	}

	/** Set our position (S coordinate). */
	protected void setS(int pos) {
		position = pos;
	}

	/** Get our S coordinate. */
	protected int getS() {
		return position;
	}

	/** Return true if this row contains boxes. */
	protected boolean isBoxRow() {
		return (boxes.length>0);
	}

	/** Calculate the thickness of the row when it is a box row.
	 * The thickness is the maximum of the thickness of all boxes
	 * in the row.
	 */
	protected int calculateThickness() {
		int Smax = 0;
		for (int b=0; b<boxes.length; b++) {
			int S = boxes[b].getThickness();
			if (S>Smax)
				Smax = S;
		}
		thickness = Smax;
		return thickness;
	}

	/** Get the previously calculated thickness. */
	protected int getThickness() {
		return thickness;
	}

	/** Get the length of the row. */
	protected int getLength() {
		return length;
	}

	/** Clear the row ordering info. */
	protected void clearRowOrder() {
		rowOrderMin = 0;
		rowOrderMax = 0;
	}

	/** Set the R position of the boxes within this row,
	 * assuming minimal spacing between boxes.
	 * This gives us a minimum length for the row.
	 */
	protected void normalize() {
		if (boxes.length==0) {
			length = 0;
			return;
		}
		int r = 0;
		for (int b=0; b<boxes.length; b++) {
			Box box = boxes[b];
			box.setR(r);
			r += layout.interBoxSpace + box.getLength();
		}
		length = r - layout.interBoxSpace;
	}

	/** Spread out the boxes in a row to minimize line lengths.
	 * This changes the R position of the boxes within the row.
	 */
	protected void balanceAndSpread(int side) {
		if (!isBoxRow())
			return;		/* don't do this if no boxes */
		balance(side);
		spread(side);
	}

	/** Move the row as a whole in the R dimension to minimize the
	 * force on the specified connectors.
	 */
	protected void balance(int side) {
		int d = getBalanceDistance(side);
		if (d<=0)
			return;	/* no moving to be done. */
		for (int b=0; b<boxes.length; b++)
			boxes[b].moveR(d);
	}

	/** Get the distance needed to move the row to the optimal
	 * balanced position. */
	protected int getBalanceDistance(int side) {
		int[] forcesums = new int[2];
			/* [0] is force sum, [1] is count */
		for (int b=0; b<boxes.length; b++)
			boxes[b].addNetForce(side,forcesums);
		if (forcesums[1]==0)
			return 0;		/* nothing added */
		int dist = forcesums[0]/forcesums[1];
		if (dist<=0)
			return 0;		/* no need to move */
		int maxDist = layout.maxRowLength - length;
			/* we can't move more than this, or the other end
			 * of the row will move outside the allowed area. */
		if (dist>maxDist)
			dist = maxDist;	/* limit how much we move the row */
		return dist;
	}

	/** Spread out the boxes within the row to minimize the force on
	 * the specified connectors of each box.
	 * First we go through all the boxes starting at the low-R end of 
	 * the row and see if we can pull boxes toward lower R; then we
	 * go through the boxes in the other direction, from the high-R end,
	 * and see if we can pull any boxes toward higher R. 
	 */
	protected void spread(int side) {
		/* Clear out the positionSet flags. */
		for (int b=0; b<boxes.length; b++)
			boxes[b].setPositionSet(false);

		int floor = 0;	/* This is the lowest allowed value for a box */
		/* Start at the low-R end and see if we can move any boxes
		 * to a lower R value. */
		int forcesums[] = new int[2];
		for (int b=0; b<boxes.length; b++) {
			forcesums[0] = 0;
			forcesums[1] = 0;
			Box box = boxes[b];
			box.addNetForce(side,forcesums);
			if (forcesums[1]==0) {
				/* No connections force the position of this
				 * box, so it can go anywhere.  We raise the
				 * floor as if we moved this box all the way
				 * to the lowest R, but we leave the
				 * positionSet flag false so that we can come
				 * back later and move it again.
				 */
				 floor+= box.getLength() + layout.interBoxSpace;
				 continue;
			}
			int dist = forcesums[0]/forcesums[1];
			if (dist>=0) {
				/* This box wants to move to higher R, but
				 * we don't do that here.  Since we're not
				 * going to move this box, we can't move any
				 * of the following boxes to a lower R, so
				 * we quit from this loop.
				 */
				break;
			}
			int maxDist = floor - box.getR();
				/* this is a negative number */
			if (dist<maxDist)
				dist = maxDist;	/* can't move more than this */
			box.moveR(dist);
			floor = box.getR() + box.getLength() +
					layout.interBoxSpace;
			box.setPositionSet(true);
				/* the position of this box is now fixed. */
			int tfloor = box.getR();
			/* Now we go back down through the boxes we have
			 * already processed, in case there were any with
			 * no force, which we moved all the way down.
			 * We now move those all the way back up to this box.
			 */
			for (int j=b-1; j>=0; j--) {
				Box boxj = boxes[j];
				if (boxj.isPositionSet())
					break;
				tfloor -= boxj.getLength() +
						layout.interBoxSpace;
				boxj.setR(tfloor);
				box.setPositionSet(true);
			}
		}

		/* Now we do exactly the same thing, except we start from
		 * the other end, using a ceiling instead of a floor,
		 * and trying to move the boxes to the higher end if we can.
		 */
		int ceil = layout.maxRowLength;
			/* This is the highest allowed value for a box */
		/* Start at the high-R end and see if we can move any boxes
		 * to a higher R value. */
		for (int b=boxes.length-1; b>=0; b--) {
			forcesums[0] = 0;
			forcesums[1] = 0;
			Box box = boxes[b];
			box.addNetForce(side,forcesums);
			if (forcesums[1]==0) {
				/* No connections force the position of this
				 * box, so it can go anywhere.  We lower the
				 * ceiling as if we moved this box all the way
				 * to the highest R, but we leave the
				 * positionSet flag false so that we can come
				 * back later and move it again.
				 */
				 ceil -= box.getLength() + layout.interBoxSpace;
				 continue;
			}
			int dist = forcesums[0]/forcesums[1];
			if (dist<=0) {
				/* This box wants to move to lower R, but
				 * we don't do that here.  Since we're not
				 * going to move this box, we can't move any
				 * of the following boxes to a higher R, so
				 * we quit from this loop.
				 */
				break;
			}
			int maxDist = ceil - box.getR() - box.getLength();
			if (dist>maxDist)
				dist = maxDist;	/* can't move more than this */
			box.moveR(dist);
			ceil = box.getR() - layout.interBoxSpace;
			box.setPositionSet(true);
				/* the position of this box is now fixed. */
			int tceil = box.getR() + box.getLength() +
					layout.interBoxSpace;
			/* Now we go back up through the boxes we have
			 * already processed, in case there were any with
			 * no force, which we moved all the way up.
			 * We now move those all the way back down to this box.
			 */
			for (int j=b+1; j<boxes.length; j++) {
				Box boxj = boxes[j];
				if (boxj.isPositionSet())
					break;
				boxj.setR(tceil);
				tceil += boxj.getLength() +
						layout.interBoxSpace;
				box.setPositionSet(true);
			}
		}
	}

	/** Set the channel size for the row.
	 * This sets thickness for a channel row.
	 */
	protected void setChannelSize() {
		tracksVector = new Vector();

		/* If the row has boxes, then it's not a routing row. */
		if (isBoxRow())
			return;
		/* Calculate the span of each net in this row. */
		for (int n=0; n<nets.length; n++)
			nets[n].calculateSpan();
		
		/* Sort the nets based on span starting point. */
		Net.sortBySpan(nets);

		/* Set the routing angle for each net. */
		for (int n=0; n<nets.length; n++)
			nets[n].calculateRoutingAngle();
		
		for (int n=nets.length-1; n>=0; n--) {
			Net net = nets[n];
			if (!net.needsTrackAssignment())
				continue;
			/* Find a track into which the span of this net fits. */
			Track track = findTrack(net.minR, net.maxR, net.angle);
			if (track==null) {
				/* Couldn't find a track, make another one. */
				track = new Track(this,tracksVector.size());
				tracksVector.addElement(track);
			}
			net.setTrack(track);
			track.addNet(net);
		}
		for (int n=0; n<nets.length; n++) {
			Net net = nets[n];
			if (net.track!=null)
				continue;	/* already assigned */
			if (net.connectors.length<=1)
				continue;	/* no routing needed */
			Track track = findTrack(net.minR, net.maxR, net.angle);
			if (track==null) {
				track = new Track(this,tracksVector.size());
				tracksVector.addElement(track);
			}
			net.setTrack(track);
			track.addNet(net);
		}
		thickness = layout.trackSpace * (tracksVector.size()+1);
	}

	/** Find a track into which the specified span will fit. */
	protected Track findTrack(int minR, int maxR, int angle) {
		if (layout.useGreedyRoute)
			return findTrackGreedy(minR, maxR, angle);
		else
			return findTrackLast(minR, maxR, angle);
	}

	/** Find a track into which the specified span will fit,
	 * using a greedy algorithm.
	 */
	protected Track findTrackGreedy(int minR, int maxR, int angle) {
		for (int t=0; t<tracksVector.size(); t++) {
			Track track = (Track)tracksVector.elementAt(t);
			if (track.used(minR,maxR)<0)
				return track;
		}
		return null;	/* no tracks available */
	}

	/** Find a track into which the specified span will fit,
	 * using a generous algorithm.
	 * Find the last row in which the span does not fit, then
	 * return the row after that.
	 */
	protected Track findTrackLast(int minR, int maxR, int angle) {
		Track lastAvailable = null;
		for (int t=tracksVector.size()-1; t>0; t--) {
			Track track = (Track)tracksVector.elementAt(t);
			int trackAngle = track.used(minR,maxR);
			if (trackAngle<0)
				lastAvailable = track;
			else
				if (trackAngle==angle)
					break;
			/* If the angles are different, continue looking. */
		}
		return lastAvailable;	/* may be null */
	}

	/** Do the channel route for this row. */
	protected void route() {
		if (isBoxRow())
			return;		/* no routing in box rows */

		/* Set the S coordinate for all tracks in this channel. */
		for (int t=0; t<tracksVector.size(); t++) {
			Track track = (Track)tracksVector.elementAt(t);
			track.setS(getS()+layout.trackSpace*(t+1));
		}

		/* Route all of the nets. */
		for (int n=0; n<nets.length; n++)
			nets[n].route();
	}

	/** True if this row contains the specified point. */
	protected boolean contains(int x, int y) {
		int xMin, xMax, yMin, yMax;
		if (layout.verticalRows) {
			xMin = position;
			xMax = position+thickness;
			yMin = 0;
			yMax = layout.maxRowLength;
		} else {
			yMin = position;
			yMax = position+thickness;
			xMin = 0;
			xMax = layout.maxRowLength;
		}
		if (x<xMin || y<yMin)
			return false;
		if (x>xMax || y>yMax)
			return false;
		return true;
	}

	/** Find the text at the specified point. */
	protected LayoutHitText getHitText(int x, int y) {
		for (int b=0; b<boxes.length; b++) {
			if (boxes[b].contains(x,y)) {
				return boxes[b].getHitText(x,y);
			}
		}
		return null;	/* not in any box in this row */
	}

	/** Dump a row for debugging. */
	public void dump(PrintWriter f, String prefix, boolean details) {
		f.println(prefix+"Row: "+rowNumber);
		if (!details)
			return;
		f.println(prefix+" position="+position+
				", thickness="+thickness+
				", length="+length);
		Box.dump(f,boxes,prefix+" ",false);
		Net.dump(f,nets,prefix+" ",false);
	}

	/** Dump an array of rows. */
	public static void dump(PrintWriter f, Row[] rows, String prefix,
						boolean details) {
		f.print(prefix);
		f.print("Rows: ");
		if (rows==null) {
			f.println("null");
			return;
		}
		f.println(rows.length);
		for (int r=0; r<rows.length; r++) {
			rows[r].dump(f,prefix+" ["+r+"] ",details);
		}
	}
}

/* end */
