/* DateSpec.java
 *
 * Jim McBeath, April 15, 2002
 */

package jimmc.util;

import java.util.Date;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.text.ParseException;
import java.text.ParsePosition;
import java.text.SimpleDateFormat;

/** A representation of date or date/time information with specified precision.
 * This can encode year/month/day/hour/minute/second/fractionalSecond,
 * or any less precise subset.
 * There are two primary differences between this and Date or Timestamp:
 * <ol>
 * <li>DateSpec allows representing partial date information, thus
 *     can distinguish between "2002", "Jan, 2002" and "Jan 1, 2002".
 * <li>DateSpec does not encode the time zone.  All times are local times.
 * </ol>
 */
public class DateSpec implements Cloneable, Comparable, Serializable {
	/** Our date. */
	protected Date date;
	//We used to extend the Date class, but then various other classes
	//(such as JTable) thought they knew how to display us.
	//They displayed by formatting it themselves, but we want them to
	//display by calling our toString method.

	/** The precision of our date. */
	protected int precision;
		/** No data specified. */
		public final static int PRECISION_NONE = 0;
	 	/** Precision to the year. */
		public final static int PRECISION_YEAR = 1;
	 	/** Precision to the month. */
		public final static int PRECISION_MONTH = 2;
	 	/** Precision to the day. */
		public final static int PRECISION_DAY = 3;
	 	/** Precision to the hour. */
		public final static int PRECISION_HOUR = 4;
	 	/** Precision to the minute. */
		public final static int PRECISION_MINUTE = 5;
	 	/** Precision to the second. */
		public final static int PRECISION_SECOND = 6;
	 	/** Precision to the fractional second. */
		public final static int PRECISION_FRACTION = 7;

	/** A cached of date formats, indexed by precision value. */
	static SimpleDateFormat[] dateFormats;

	/** The format for this instance. */
	protected SimpleDateFormat dateFormat;

	/** Create a DateSpec with no data. */
	public DateSpec() {
		date = new Date(0);
		setPrecision(PRECISION_NONE);
	}

	/** Create a DateSpec from a Date and precision. */
	public DateSpec(Date date, int precision) {
		this.date = new Date(date.getTime());	//make a copy
		if (precision<PRECISION_NONE || precision>PRECISION_FRACTION) {
			String msg = "precision "+precision;
			throw new IllegalArgumentException(msg);
		}
		setPrecision(precision);
	}

	/** Create a DateSpec from a string. */
	public DateSpec(String s) {
		if (s==null)
			s="";
		s = s.trim();
		int sLen = s.length();
		if (sLen==0) {
			date = new Date(0);
			setPrecision(PRECISION_NONE);
			return;
		}
		ParsePosition pos = new ParsePosition(0);
		for (int p=PRECISION_FRACTION; p>=PRECISION_YEAR; p--) {
			SimpleDateFormat dateFmt = getDateFormat(p);
			pos.setIndex(0);
			Date d = (Date)dateFmt.parseObject(s,pos);
			int x = pos.getIndex();
			if (x>=sLen && d!=null) {
				//Parsed it all
				date = new Date(d.getTime());
				setPrecision(p);
				return;
			}
		}
		//Unable to parse any format
		String msg = "Unable to parse: "+s;	//TBD i18n
		throw new IllegalArgumentException(msg);
	}

	/** Create a DateSpec from a Date.
	 * The precision is set to PRECISION_SECOND.
	 */
	public DateSpec(Date date) {
		this(date,PRECISION_SECOND);
	}

	/** Get our date. */
	public Date getDate() {
		return date;
	}

	/** Set our precision. */
	public void setPrecision(int precision) {
		this.precision = precision;
		setDateFormat();
	}

	/** Get our precision. */
	public int getPrecision() {
		return precision;
	}

	/** Get our canonical date format according to the precision. */
	protected static SimpleDateFormat getDateFormat(int precision) {
		if (precision<PRECISION_NONE || precision>PRECISION_FRACTION) {
			String msg = "precision "+precision;
			throw new IllegalArgumentException(msg);
		}
		if (dateFormats==null) {
			//One-time initialization of our formats
			dateFormats= new SimpleDateFormat[PRECISION_FRACTION+1];
			dateFormats[PRECISION_YEAR] =
				new SimpleDateFormat("yyyy");
			dateFormats[PRECISION_MONTH] =
				new SimpleDateFormat("yyyy-MM");
			dateFormats[PRECISION_DAY] =
				new SimpleDateFormat("yyyy-MM-dd");
			dateFormats[PRECISION_HOUR] =
				new SimpleDateFormat("yyyy-MM-dd HH");
			dateFormats[PRECISION_MINUTE] =
				new SimpleDateFormat("yyyy-MM-dd HH:mm");
			dateFormats[PRECISION_SECOND] =
				new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
			dateFormats[PRECISION_FRACTION] =
				new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.S");
		}
		return dateFormats[precision];
	}

	/** Set our canonical date format according to the precision. */
	protected void setDateFormat() {
		dateFormat = getDateFormat(precision);
	}

	/** Convert to a String.
	 * The format used is the same as the standard SQL format,
	 * <i>YYYY-MM-DD HH:MM:SS.F</i>, or a truncated version of
	 * that format when less precision has been specified.
	 */
	public String toString() {
		if (precision==PRECISION_NONE)
			return "";
		return dateFormat.format(date);
	}

	/** Create a DateSpec from a string.
	 * Parses the same format of string as is created by
	 * {@link #toString}.
	 */
	public static DateSpec valueOf(String s) {
		if (s==null || s.trim().equals(""))
			return null;	//no data
		return new DateSpec(s);
	}

    //The Clonable interface
	public Object clone() {
		DateSpec ds = null;
		try {
			ds = (DateSpec)super.clone();
			ds.date = new Date(this.date.getTime());
			ds.setPrecision(this.precision);
		} catch (CloneNotSupportedException ex) {}
		return ds;
	}
    //End Clonable interface

    //The Comparable interface
	public int compareTo(Object o) {
		int d = date.compareTo(o);
		if (d!=0)
			return d;
		if (o instanceof DateSpec)
			return precision - ((DateSpec)o).getPrecision();
		return 0;
	}
    //End Comparable interface

    //The Serializable interface
	private void writeObject(ObjectOutputStream s) throws IOException {
		s.writeLong(date.getTime());
		s.writeInt(precision);
	}

	private void readObject(ObjectInputStream s) 
			throws IOException, ClassNotFoundException {
		date = new Date(s.readLong());
		setPrecision(s.readInt());
	}
    //End Serializable interface
}

/* end */
